
intro = """pdfriend: a command line utility for easily modifying PDF files
    usage: pdfriend [command] [arguments?] [-d|--debug?] (note that options in [] are required and options in [?] are not). Use -d or --debug for more detailed error messages.
    the following commands are available:
    """

outro = """use pdfriend help [command] to get the instructions for particular commands"""

help_blurbs = {
    "version":"""pdfriend version | -v | --version
    prints the current version of pdfriend
    """,
    "help": """pdfriend help [command?]
        display help message. If given a command, it will only display the help message for that command.
    """,
    "merge": """pdfriend merge [filename1] [filename2?] ... [-o|--outfile outfile?=pdfriend_output.pdf] [-q|--quality quality?=100]
        merge the given files into one pdf. It can handle multiple pdfs, as well convert and merge png and jpg images. Glob patterns are also supported. You can specify the output filename using the -o or --outfile flag, otherwise it defaults to pdfriend_output.pdf. You can also specify the quality when images are converted to pdfs via the -q or --quality flag. It's an integer going from 0 to 100, 100 is no lossy compression and 0 is full lossy compression.

    examples:
        pdfriend merge pdf1.pdf img.png pdf2.pdf -o merged.pdf
            merges all of those into merged.pdf, preserving the quality of img.png
        pdfriend merge folder_name/* -o merged.pdf -q 50
            merges all files in directory folder_name into merged.pdf and compresses the images by 50%.
        pdfriend merge pdf1.pdf folder_name/* img.jpg pdf2.pdf -o apricot.pdf
            merges every file given, including all files in folder_name, into apricot.pdf
    """,
    "split": """pdfriend split [filename] [pages] ... [-o|--outfile outfile?=pdfriend_output]
        split the given file at the given points. Every point is included in the part after, not before it.

        examples:
            pdfriend split in.pdf 5 -o parts
                splits in.pdf into one part with pages 1-4 and another with 5-end and puts them in a directory named parts
            pdfriend split input.pdf 4,7
                splits input.pdf into 3 parts, one has pages 1-3, another 4-6 and another 7-end and puts them in a directory named pdfriend_output
            pdfriend split thing.pdf 8-11
                splits thing.pdf into 4 parts, one has pages 1-7, another has page 8, the other page 9, the other page 10, and the other pages 11-end and puts them in a directory named pdfriend_output
            pdfriend split infile.pdf all -o pages
                splits infile.pdf into individual pages and puts them in a directory named pages
    examples:
        pdfriend merge pdf1.pdf img.png pdf2.pdf -o merged.pdf
            merges all of those into merged.pdf, preserving the quality of img.png
        pdfriend merge folder_name/* -o merged.pdf -q 50
            merges all files in directory folder_name into merged.pdf and compresses the images by 50%.
        pdfriend merge pdf1.pdf folder_name/* img.jpg pdf2.pdf -o apricot.pdf
            merges every file given, including all files in folder_name, into apricot.pdf
    """,
    "edit": """pdfriend edit [filename]
        edit the selected file in place, using a set of subcommands. After launching the edit shell, you can type h or help to list the subcommands.
    """,
    "invert": """pdfriend invert [filename] [-o|--outfile outfile?=pdfriend_output.pdf] [-i|--inplace?]
        create a PDF file with the pages of the input file, but in inverted order. Adding -i or --inplace will make it so the input file is modified, instead of creating a new one.

        examples:
            pdfriend invert puppy.pdf -o puppy-inv.pdf
                inverts the pages of puppy.pdf and saves to puppy-inv.pdf
            pdfriend invert kitty.pdf -i
                inverts the pages of kitty.pdf
    """,
    "clear": """pdfriend clear
        clears the pdfriend cache.
    """,
    "swap": """pdfriend swap [filename] [page_0] [page_1] [-o|--outfile?=pdfriend_output.pdf] [-i|--inplace?]
        swaps the specified pages in the PDF file. Adding -i or --inplace will make it so the input file is modified, instead of creating a new one.

        examples:
            pdfriend swap notes.pdf 1 3 -i
                swaps pages 1 and 3 in notes.pdf (modifies the file)
            pdfriend swap templ.pdf 6 3 -o new-templ.pdf
                swaps pages 6 and 3 and saves to new-templ.pdf
            pdfriend swap k.pdf 2 9
                swaps pages 2 and 9 and saves to pdfriend_output.pdf
    """,
    "remove": """pdfriend remove [filename] [pages] [-o|--outfile?=pdfriend_output.pdf] [-i|--inplace?]
        removes specified pages from the PDF file. Adding -i or --inplace will make it so the input file is modified, instead of creating a new one.

        examples:
            pdfriend remove input.pdf 6
                removes page 6 and saves to pdfriend_output.pdf
            pdfriend remove input.pdf 5,6,9 -o out.pdf
                removes pages 5,6,9 and saves to out.pdf
            pdfriend remove input.pdf 3-7 -o out
                removes pages 3 through 7 (INCLUDING 7) and saves to out.pdf
            pdfriend remove input.pdf 13,2,4-7,1 -i
                removes pages 1,2,4,5,6,7,13 from input.pdf (modifies the file)
    """,
    "weave": """pdfriend weave [filename_0] [filename_1] [-o|--outfile?=pdfriend_output.pdf]
        combines two PDFs such that the first fills the odd pages and the second the even pages of the output.

        examples:
            pdfriend weave inf0.pdf inf1.pdf
                weaves the two PDFs and saves the output to pdfriend_output.pdf
            pdfriend weave k.pdf l.pdf -o weaved.pdf
                weaves the two PDFs and saves the output to weaved.pdf
    """,
    "encrypt": """pdfriend encrypt [filename] [-o|--outfile?=pdfriend_output.pdf] [-i|--inplace?]
        creates an encrypted PDF file using a provided password. Adding -i or --inplace will make it so that the file itself is encrypted.

        examples:
            pdfriend encrypt not-sus.pdf -i
                encrypts not-sus.pdf in-place. Make sure you remember the password, as it will be overwritten!
            pdfriend encrypt balance.pdf -o balance-encrypted.pdf
                encrypts balance.pdf and saves to balance-encrypted.pdf.
            pdfriend encrypt acct.pdf
                encrypts acct.pdf and saves to pdfriend_output.pdf.
    """,
    "decrypt": """pdfriend decrypt [filename] [-o|--outfile?=pdfriend_output.pdf] [-i|--inplace?]
        decrypts an encrypted PDF file using a provided password. Adding -i or --inplace will make it so that the file itself is decrypted. If the file is not encrypted, it will just be copied.

        examples:
            pdfriend decrypt not-sus.pdf -i
                decrypts not-sus.pdf in-place.
            pdfriend decrypt balance.pdf -o balance-decrypted.pdf
                decrypts balance.pdf and saves to balance-decrypted.pdf.
            pdfriend decrypt acct.pdf
                decrypts acct.pdf and saves to pdfriend_output.pdf.
    """,
    "metadata": """pdfriend metadata [filename] [--get key?] [--set key_val_pairs?] [--pop keys?]
        manages PDF metadata. Using no extra flags, it will print the key-value pairs. You can use --get to print the value of a specific key and --set to set values for keys, or --pop to delete them.

        examples:
            pdfriend metadata some.pdf
                prints the metadata of some.pdf.
            pdfriend metadata thing.pdf --get /Author
                prints the name of the author of the PDF, if that field has been set.
            pdfriend metadata stolen.pdf --set /Author=me
                sets the author of stolen.pdf to "me". BEWARE: This will overwrite the PDF, unlike most of the other pdfriend commands.
            pdfriend metadata cnp.pdf --set  "/Title=Crime And Punishment"
                sets the title to that. Note that you need the quotes here, else your shell will interpret the words as different arguments.
            pdfriend metadata phys1.pdf --set "/Title=University Physics with Modern Physics,/Author=H. Young and R. Freedman"
                sets the author of phys1.pdf to "H. Young and R. Freedman" and its title to you-know-what.
            pdfriend metadata embarassing_fanfic.pdf --pop /Author
                removes (!) the PDF's author field.
            pdfriend metadata mystery.pdf --pop /Author,/Producer
                removes the PDF's author and producer fields.
    """
}


def help(command: str):
    try:
        print(help_blurbs[command])
    except KeyError:
        print(intro)
        print("\n".join([f"    {blurb}" for blurb in help_blurbs]), "\n")
        print(outro)



import argparse
from pdfriend.classes.config import Config
from pdfriend.classes.platforms import Platform
import pdfriend.classes.cmdparsers as cmdparsers
import pdfriend.classes.exceptions as exceptions
import pdfriend.commands as commands
import pdfriend.utils as utils


def main():
    parser = argparse.ArgumentParser(add_help=False)

    parser.add_argument("commands", type=str, nargs="*")

    parser.add_argument("-h", "--help", action="store_true")
    parser.add_argument("-v", "--version", action="store_true")
    parser.add_argument("-d", "--debug", action="store_true")

    parser.add_argument("-o", "--outfile", type=str, default="pdfriend_output")
    parser.add_argument("-i", "--inplace", action="store_true")
    parser.add_argument("-q", "--quality", type=int, default=100)

    parser.add_argument("--get", type=str, default=None)
    parser.add_argument("--set", type=str, default=None)
    parser.add_argument("--pop", type=str, default=None)

    args = parser.parse_args()

    Platform.Init()
    Config.Debug = args.debug

    command = ""
    if len(args.commands) > 0:
        command = args.commands[0]

    cmd_parser = cmdparsers.CmdParser(command, args.commands[1:])

    try:
        if command == "version" or args.version:
            print(commands.version())
        elif command == "help" or args.help:
            command_to_display = cmd_parser.next_str_or(None)
            commands.help(command_to_display)
        elif command == "merge":
            if len(cmd_parser.args) == 0:
                print("You need to specify at least one file or pattern to be merged")
                return

            commands.merge(cmd_parser.args, args.outfile, args.quality)
        elif command == "edit":
            infile = cmd_parser.next_str("filename")
            commands.edit(infile)
        elif command == "invert":
            infile = cmd_parser.next_str("filename")

            if args.inplace:
                args.outfile = infile

            commands.invert(infile, args.outfile)
        elif command == "swap":
            infile = cmd_parser.next_str("filename")
            page_0 = cmd_parser.next_int("page_0")
            page_1 = cmd_parser.next_int("page_1")

            if args.inplace:
                args.outfile = infile

            commands.swap(infile, page_0, page_1, args.outfile)
        elif command == "clear":
            commands.clear()
        elif command == "remove":
            infile = cmd_parser.next_str("filename")
            slice = cmd_parser.next_str("pages")

            if args.inplace:
                args.outfile = infile

            commands.remove(infile, slice, args.outfile)
        elif command == "weave":
            infile_0 = cmd_parser.next_str("filename_0")
            infile_1 = cmd_parser.next_str("filename_1")

            commands.weave(infile_0, infile_1, args.outfile)
        elif command == "split":
            infile = cmd_parser.next_str("filename")
            slice = cmd_parser.next_str("pages")

            commands.split(infile, slice, args.outfile)
        elif command == "encrypt":
            infile = cmd_parser.next_str("filename")

            if args.inplace:
                args.outfile = infile

            commands.encrypt(infile, args.outfile)
        elif command == "decrypt":
            infile = cmd_parser.next_str("filename")

            if args.inplace:
                args.outfile = infile

            commands.decrypt(infile, args.outfile)
        elif command == "metadata":
            infile = cmd_parser.next_str("filename")

            commands.metadata(infile, args.get, args.set, args.pop)
        else:
            print(f"command \"{command}\" not recognized")
            print("use pdfriend help for a list of the available commands")
    except exceptions.ExpectedError as e:
        print(e)
    except Exception as e:
        utils.print_unexpected_exception(e, Config.Debug)



import pdfriend.classes.wrappers as wrappers


def remove(infile: str, slice_str: str, outfile: str):
    pdf = wrappers.PDFWrapper.Read(infile)

    for page in pdf.slice(slice_str):
        pdf.pop_page(page)

    pdf.write(outfile)



import pdfriend.classes.wrappers as wrappers

def weave(infile_0: str, infile_1: str, outfile: str):
    pdf_0 = wrappers.PDFWrapper.Read(infile_0)
    pdf_1 = wrappers.PDFWrapper.Read(infile_1)

    weaved_pdf = pdf_0.weave_with(pdf_1)

    weaved_pdf.write(outfile)



import pdfriend.classes.wrappers as wrappers
import pdfriend.classes.exceptions as exceptions
import pdfriend.classes.cmdparsers as cmdparsers
import pdfriend.classes.info as info
from pdfriend.classes.platforms import Platform
from pdfriend.classes.config import Config
import pdfriend.utils as utils
import pathlib
import re


whitespace_pattern = re.compile(r"\s+")

def parse_command_string(command_string: str) -> list[str]:
    return re.split(whitespace_pattern, command_string)

command_info = {
    "help": info.CommandInfo("help", "h", """[command?]
    display help message. If given a command, it will only display the help message for that command.

    examples:
        help rotate
            displays the help blurb for the rotate command
        help exit
            displays the help blurb for the exit command
    """),
    "exit": info.CommandInfo("exit", "e", """
    exits the edit mode
    """),
    "rotate": info.CommandInfo("rotate", "r", """[page_numbers] [angle]
    rotates page clockwise with the given numbers (starting from 1) by the given angle (in degrees). Can use negative angles to rotate counter-clockwise. DO NOT put extra spaces between the page numbers!

    examples:
        r 34 1.2
            rotates page 34 clockwise by 1.2 degrees
        r 1,3,8 -4
            rotates pages 1,3 and 8 counter-clockwise by 4 degrees
        r 3-18 90
            rotates pages 3 through 18 (INCLUDING 18) clockwise by 90 degrees
        r 1,13,5-7,2 54
            rotates pages 1,2,5,6,7,13 clockwise by 54 degrees
        r all -90
            rotates all pages counter-clockwise by 90 degrees
    """),
    "delete": info.CommandInfo("delete", "d", """[page_numbers]
    deletes all specified pages. DO NOT put extra spaces between the page numbers!

    examples:
        d 7
            deletes page 7
        d 4,8,1
            deletes pages 1, 4 and 8
        d 6-66
            deletes pages 6 through 66 (INCLUDING 66)
        d 4,17,3-6
            deletes pages 3,4,5,6 and 17
    """),
    "swap": info.CommandInfo("swap", "s", """[page_0] [page_1]
    swaps page_0 and page_1.
    """),
    "move": info.CommandInfo("move", "m", """[source] [destination]
    move source to BEFORE destination, taking its place.

    examples:
        m 3 17
            moves page 3 to page 17, pushing back the pages from 17 onward
        m 83 1
            moves page 83 to the beginning of the document
    """),
    "push": info.CommandInfo("push", "p", """[pages] [offset]
    pushes the specified pages by offset pages (offset can be negative).

    examples:
        p 3 7
            moves page 3 to 7 pages down, i.e. to page 10.
        p 4,9,2 1
            moves pages 2,4,9 by 1 page.
        p 5-8 -3
            moves pages 5,6,7,8 to 3 pages BACK.
        p 5,6,90-94 5
            moves pages 5,6,90,91,92,93,94 to be 5 pages down.
        p -5 4
            moves pages 1,2,3,4,5 to be 4 pages down.
        p 67- -7
            move pages from 67 to the end of the PDF to be 7 pages back.
        p 70- 5
            FAILS. 70- includes the end of the PDF, and you can't move that further down.
    """),
    "undo": info.CommandInfo("undo", "u", """[number?]
    undo the previous [number] commands.

    examples:
        u
            undoes the previous command
        u 3
            undoes the previous 3 commands
        u all
            undoes all commands issued this session (reverts document fully)
    """),
}

command_info_by_shorts = {cmd.short: cmd for name, cmd in command_info.items()}

def print_help(subcommand: str | None = None):
    if subcommand is None:
         print("pdfriend edit shell for quick changes. Commands:")
         for command, info in command_info.items():
             print(f"{command} (short: {info.short})")

         print("use h [command] to learn more about a specific command")
         return

    sub_info = None
    if subcommand in command_info_by_shorts:
        sub_info = command_info_by_shorts[subcommand]
    elif subcommand in command_info:
        sub_info = command_info[subcommand]
    else:
        raise exceptions.ExpectedError(f"command \"{subcommand}\" does not exist")

    print(f"{sub_info.name}|{sub_info.short} {sub_info.descr}")



def run_edit_command(pdf: wrappers.PDFWrapper, args: list[str]):
    no_command_msg = "No command specified! Type h or help for a list of the available commands"
    if len(args) == 0:
        raise exceptions.ExpectedError(no_command_msg)

    command = args[0]
    if command == "":
        raise exceptions.ExpectedError(no_command_msg)

    short = ""; long = ""
    if command in command_info_by_shorts:
        short = command
        long = command_info_by_shorts[command].name
    elif command in command_info:
        long = command
        short = command_info[command].short
    else:
        raise exceptions.ExpectedError(f"command \"{command}\" does not exist")

    cmd_parser = cmdparsers.CmdParser(long, args[1:])

    if short == "h":
        subcommand = cmd_parser.next_str_or(None)
        print_help(subcommand)

        # this is to prevent rewriting the file and appending
        # the command to the command stack
        raise exceptions.ShellContinue()
    if short == "e":
        raise exceptions.ShellExit()
    if short == "r":
        pages = cmd_parser.next_typed("PDF slice", lambda s: pdf.slice(s), "pages")
        angle = cmd_parser.next_float("angle")

        if len(pages) == 0:
            return
        # the slice is sorted, so if any pages are out of range, it'll
        # either be the first or the last one, probably the last
        pdf.raise_if_out_of_range(pages[-1])
        pdf.raise_if_out_of_range(pages[0])

        for page in pages:
            pdf.rotate_page(page, angle)
    if short == "d":
        pages = cmd_parser.next_typed("PDF slice", lambda s: pdf.slice(s))

        for page in pages:
            pdf.pop_page(page)
    if short == "s":
        page_0 = cmd_parser.next_int("page_0")
        pdf.raise_if_out_of_range(page_0)
        page_1 = cmd_parser.next_int("page_1")
        pdf.raise_if_out_of_range(page_1)

        pdf.swap_pages(page_0, page_1)
    if short == "m":
        source = cmd_parser.next_int("source")
        pdf.raise_if_out_of_range(source)
        destination = cmd_parser.next_int("destination")
        pdf.raise_if_out_of_range(destination)

        page = pdf.pages.pop(source - 1)
        pdf.pages.insert(destination - 1, page)
    if short == "p":
        pages = cmd_parser.next_typed("PDF slice", lambda s: pdf.slice(s), "pages")
        offset = cmd_parser.next_int("offset")

        last_page_before = pages[-1]
        last_page_after = last_page_before + offset

        if last_page_after > pdf.len(): # only check last page, as the slice is sorted
            raise exceptions.ExpectedError(
                f"can't move page {last_page_before} to {last_page_after}, as it's outside the PDF (number of pages: {pdf.len()})"
            )

        if offset > 0:
            pages = pages[::-1]

        for page in pages:
            p = pdf.pages.pop(page - 1)
            pdf.pages.insert(page + offset - 1, p)
    if short == "u":
        # arg will be converted to int, unless it's "all". Defaults to 1
        num_of_commands = cmd_parser.next_typed_or(
            "int or \"all\"", lambda s: s if s == "all" else int(s),
            1 # default value
        )

        raise exceptions.ShellUndo(num_of_commands)


def edit(infile: str):
    pdf = wrappers.PDFWrapper.Read(infile)
    command_stack = []

    # backup the file, because it will be overwritten
    backup_path = pdf.backup(infile)
    print(f"editing {infile}\nbackup created in {backup_path}")

    while True:
        try:
            args = parse_command_string(input(""))
            run_edit_command(pdf, args)
            command_stack.append(args)

            pdf.write(infile) # overwrites the file!
        except (KeyboardInterrupt, exceptions.ShellExit):
            return
        except exceptions.ShellContinue:
            continue
        except exceptions.ShellUndo as undo:
            if undo.num == "all":
                command_stack = []
            else:
                command_stack = command_stack[:-undo.num]

            pdf = wrappers.PDFWrapper.Read(backup_path)
            for args in command_stack:
                run_edit_command(pdf, args)

            pdf.write(infile)
        except exceptions.ExpectedError as e:
            print(e)
        except Exception as e:
            utils.print_unexpected_exception(e, Config.Debug)




import traceback

def print_unexpected_exception(exception: Exception, debug: bool = False):
    print("unexpected error occured:")
    if debug:
        traceback.print_exception(exception)
    else:
        print(exception)




from pdfriend.commands.version import version
from pdfriend.commands.help import help
from pdfriend.commands.merge import merge
from pdfriend.commands.invert import invert
from pdfriend.commands.remove import remove
from pdfriend.commands.swap import swap
from pdfriend.commands.edit import edit
from pdfriend.commands.clear import clear
from pdfriend.commands.weave import weave
from pdfriend.commands.split import split
from pdfriend.commands.encrypt import encrypt
from pdfriend.commands.decrypt import decrypt
from pdfriend.commands.metadata import metadata



import pdfriend.classes.exceptions as exceptions
import pdfriend.classes.wrappers as wrappers

def arg_str(arg_num: int, arg_name: str | None) -> str:
    arg_name_str = "" if arg_name is None else f" (\"{arg_name}\")"

    return f"argument {arg_num}{arg_name_str}"

class CmdParser:
    def __init__(self, cmd_name: str, args: list[str]):
        self.cmd_name = cmd_name
        self.args = args
        self.current_arg = 1

    def next_str(self, name: str | None = None):
        if len(self.args) == 0:
            raise exceptions.ExpectedError(
                f"{arg_str(self.current_arg, name)} for command \"{self.cmd_name}\" not provided"
            )

        head, tail = self.args[0], self.args[1:]
        self.args = tail
        self.current_arg += 1

        return head

    def next_str_or(self, default: str, name: str | None = None) -> str:
        try:
            return self.next_str(name)
        except Exception:
            return default

    def next_typed(self, type_name: str, type_converter, name: str | None = None):
        head = self.next_str(name)

        try:
            result = type_converter(head)
            return result
        except Exception:
            self.current_arg -= 1 # it gets incremented in next_str()
            raise exceptions.ExpectedError(
                f"value \"{head}\" of {arg_str(self.current_arg, name)} for command \"{self.cmd_name}\" could not be converted to type \"{type_name}\""
            )

    def next_typed_or(self, type_name: str, type_converter, default, name: str | None = None):
        try:
            return self.next_typed(type_name, type_converter, name)
        except Exception:
            return default

    def next_int(self, name: str | None = None) -> int:
        return self.next_typed("int", int, name)

    def next_int_or(self, default: int, name: str | None = None) -> int:
        try:
            return self.next_int(name)
        except Exception:
            return default

    def next_float(self, name: str | None = None) -> float:
        return self.next_typed("float", float, name)

    def next_float_or(self, default: float, name: str | None = None) -> float:
        try:
            return self.next_float(name)
        except Exception:
            return default

    def next_pdf_slice(self, pdf: wrappers.PDFWrapper, name: str | None = None) -> list[int]:
        return self.next_typed("PDF slice", lambda s: pdf.slice(s), name = name)

    def next_pdf_slice_or(self, pdf: wrappers.PDFWrapper, default: list[int], name: str | None = None) -> list[int]:
        try:
            return self.next_pdf_slice(pdf, name = name)
        except Exception:
            return default





import pdfriend.classes.wrappers as wrappers


def swap(infile: str, page_num_0: int, page_num_1: int, outfile: str):
    pdf = wrappers.PDFWrapper.Read(infile)

    pdf.swap_pages(page_num_0, page_num_1)

    pdf.write(outfile)
